from typing import Annotated
from sqlalchemy import create_engine, text

from admyral.action import action, ArgumentMetadata
from admyral.typings import JsonValue
from admyral.context import ctx


@action(
    display_name="Run SQL Query",
    display_namespace="Database",
    description="Run a SQL query on a database. Supported databases: PostgreSQL, MySQL, and MariaDB.",
    secrets_placeholders=["DB_URI"],
)
def run_sql_query(
    sql_query: Annotated[
        str,
        ArgumentMetadata(
            display_name="SQL Query", description="The query to run on the database"
        ),
    ],
) -> JsonValue:
    secret = ctx.get().secrets.get("DB_URI")
    db_uri = secret["uri"]

    if db_uri.startswith("mysql://"):
        db_uri = db_uri.replace("mysql://", "mysql+pymysql://")

    with create_engine(db_uri).connect() as connection:
        result = connection.execute(text(sql_query))
        if not result.returns_rows:
            return
        columns = result.keys()
        return [dict(zip(columns, row)) for row in result.fetchall()]



from typing import Annotated
import json
import time
from datetime import datetime, timedelta, UTC
import yaml

from admyral.action import action, ArgumentMetadata
from admyral.typings import JsonValue
from admyral.context import ctx


@action(
    display_name="Deserialize JSON String",
    display_namespace="Admyral",
    description="Deserializes a JSON string.",
)
def deserialize_json_string(
    serialized_json: Annotated[
        str,
        ArgumentMetadata(
            display_name="Serialized JSON String",
            description="Deserialized JSON string.",
        ),
    ],
) -> JsonValue:
    return json.loads(serialized_json)


@action(
    display_name="Serialize JSON String",
    display_namespace="Admyral",
    description="Serializes a JSON string.",
)
def serialize_json_string(
    json_value: Annotated[
        JsonValue,
        ArgumentMetadata(
            display_name="JSON",
            description="A JSON value.",
        ),
    ],
) -> str:
    return json.dumps(json_value)


@action(
    display_name="Wait",
    display_namespace="Admyral",
    description="Waits for a specified number of seconds.",
)
def wait(
    seconds: Annotated[
        int,
        ArgumentMetadata(
            display_name="Seconds",
            description="Number of seconds to wait.",
        ),
    ],
) -> None:
    time.sleep(seconds)


@action(
    display_name="Transform",
    display_namespace="Admyral",
    description="Transforms a JSON.",
)
def transform(
    value: Annotated[
        JsonValue,
        ArgumentMetadata(
            display_name="Value",
            description="A JSON value.",
        ),
    ],
) -> JsonValue:
    return value


@action(
    display_name="Get Time Interval of Last N Hours",
    display_namespace="Admyral",
    description="Get the time interval of the last N hours. A list of length 2 containing the start and end time is returned.",
)
def get_time_interval_of_last_n_hours(
    n_hours: Annotated[
        int,
        ArgumentMetadata(
            display_name="Number of Hours",
            description="Number of hours to go back.",
        ),
    ] = 1,
) -> list[str]:
    end_time = datetime.now(UTC).replace(minute=0, second=0, microsecond=0)
    start_time = (
        (end_time - timedelta(hours=n_hours)).isoformat().replace("+00:00", "Z")
    )
    return [start_time, end_time.isoformat().replace("+00:00", "Z")]


@action(
    display_name="Get Time Interval of Last N Days",
    display_namespace="Admyral",
    description="Get the time interval of the last N days. A list of length 2 containing the start and end time is returned.",
)
def get_time_interval_of_last_n_days(
    n_days: Annotated[
        int,
        ArgumentMetadata(
            display_name="Number of Days",
            description="Number of days to go back.",
        ),
    ] = 1,
) -> list[str]:
    end_time = datetime.now(UTC).replace(hour=0, minute=0, second=0, microsecond=0)
    start_time = (end_time - timedelta(days=n_days)).isoformat().replace("+00:00", "Z")
    return [start_time, end_time.isoformat().replace("+00:00", "Z")]


@action(
    display_name="Format JSON to List View String",
    display_namespace="Admyral",
    description="Converts a JSON value to a list view string (YAML format).",
)
def format_json_to_list_view_string(
    json_value: Annotated[
        JsonValue,
        ArgumentMetadata(
            display_name="JSON",
            description="A JSON value.",
        ),
    ],
) -> str:
    return yaml.dump(json_value, default_flow_style=False)


@action(
    display_name="Send to Workflow",
    display_namespace="Admyral",
    description="Sends data to a workflow.",
)
def send_to_workflow(
    workflow_name: Annotated[
        str,
        ArgumentMetadata(
            display_name="Workflow Name",
            description="The name of the workflow to send the data to.",
        ),
    ],
    payload: Annotated[
        JsonValue,
        ArgumentMetadata(
            display_name="Payload",
            description="Payload to send to the workflow.",
        ),
    ],
) -> None:
    ctx.get().send_to_workflow_sync(workflow_name, payload)


# TODO: remove with the introduction of for-loops
@action(
    display_name="Send List Elements to a Workflow",
    display_namespace="Admyral",
    description="WARNING: This action is temporary and will be deprecated as soon as For Loops are released. Sends a list of elements to a workflow.",
)
def send_list_elements_to_workflow(
    workflow_name: Annotated[
        str,
        ArgumentMetadata(
            display_name="Workflow Name",
            description="The name of the workflow to send the elements to.",
        ),
    ],
    elements: Annotated[
        list[JsonValue],
        ArgumentMetadata(
            display_name="Elements",
            description="A list of elements to send to the workflow.",
        ),
    ],
    shared_data: Annotated[
        JsonValue | None,
        ArgumentMetadata(
            display_name="Shared Data",
            description="Shared data to send to the workflow.",
        ),
    ] = None,
) -> None:
    for element in elements:
        send_to_workflow(
            workflow_name=workflow_name,
            payload={
                "element": element,
                "shared": shared_data,
            },
        )


@action(
    display_name="Split Text",
    display_namespace="Admyral",
    description="Splits a text into a list of strings.",
)
def split_text(
    text: Annotated[
        str, ArgumentMetadata(display_name="Text", description="The text to split.")
    ],
    pattern: Annotated[
        str,
        ArgumentMetadata(
            display_name="Pattern", description="The pattern to split by."
        ),
    ],
) -> list[str]:
    return text.split(pattern)



from typing import Annotated
from httpx import Client

from admyral.action import action, ArgumentMetadata
from admyral.context import ctx
from admyral.typings import JsonValue


def get_kandji_client(api_url: str, api_token: str) -> Client:
    return Client(
        base_url=f"https://{api_url}/api/v1",
        headers={
            "Authorization": f"Bearer {api_token}",
            "Content-Type": "application/json",
            "Accept": "application/json",
        },
    )


def _list_devices(client: Client) -> list[dict[str, JsonValue]]:
    offset = 0
    limit_per_page = 300

    devices = []

    while True:
        response = client.get(
            "/devices", params={"offset": offset, "limit": limit_per_page}
        )
        response.raise_for_status()

        new_devices = response.json()
        devices.extend(new_devices)

        offset += len(new_devices)
        if len(new_devices) < limit_per_page:
            break

    return devices


def _get_device_details(client: Client, device_id: str) -> dict[str, JsonValue]:
    response = client.get(f"/devices/{device_id}/details")
    response.raise_for_status()
    return response.json()


@action(
    display_name="List Devices",
    display_namespace="Kandji",
    description="List devices managed by Kandji",
    secrets_placeholders=["KANDJI_SECRET"],
)
def list_kandji_devices() -> list[dict[str, JsonValue]]:
    # https://api-docs.kandji.io/#78209960-31a7-4e3b-a2c0-95c7e65bb5f9

    secret = ctx.get().secrets.get("KANDJI_SECRET")
    api_token = secret["api_token"]
    api_url = secret["api_url"]

    with get_kandji_client(api_url, api_token) as client:
        return _list_devices(client)


@action(
    display_name="Get Device Details",
    display_namespace="Kandji",
    description="List devices managed by Kandji",
    secrets_placeholders=["KANDJI_SECRET"],
)
def get_kandji_device_details(
    device_id: Annotated[
        str,
        ArgumentMetadata(display_name="Device ID", description="The ID of the device"),
    ],
) -> list[dict[str, JsonValue]]:
    # https://api-docs.kandji.io/#efa2170d-e5f7-4b97-8f4c-da6f84ba58b5

    secret = ctx.get().secrets.get("KANDJI_SECRET")
    api_token = secret["api_token"]
    api_url = secret["api_url"]

    with get_kandji_client(api_url, api_token) as client:
        return _get_device_details(client, device_id)


@action(
    display_name="List Unencrypted Devices",
    display_namespace="Kandji",
    description="List devices managed by Kandji which don't have disk encryption enabled on all volumes",
    secrets_placeholders=["KANDJI_SECRET"],
)
def list_kandji_unencrypted_devices() -> list[dict[str, JsonValue]]:
    # https://api-docs.kandji.io/#78209960-31a7-4e3b-a2c0-95c7e65bb5f9

    secret = ctx.get().secrets.get("KANDJI_SECRET")
    api_token = secret["api_token"]
    api_url = secret["api_url"]

    with get_kandji_client(api_url, api_token) as client:
        devices = _list_devices(client)

        unencrypted_devices = []
        for device in devices:
            device_details = _get_device_details(client, device["device_id"])
            if any(volume["encrypted"] == "No" for volume in device_details["volumes"]):
                unencrypted_devices.append(device)

        return unencrypted_devices


@action(
    display_name="Get Device Apps",
    display_namespace="Kandji",
    description="List the installed apps of a Kandji managed device",
    secrets_placeholders=["KANDJI_SECRET"],
)
def get_kandji_device_apps(
    device_id: Annotated[
        str,
        ArgumentMetadata(display_name="Device ID", description="The ID of the device"),
    ],
) -> list[dict[str, JsonValue]]:
    # https://api-docs.kandji.io/#f8cd9733-89b6-40f0-a7ca-76829c6974df

    secret = ctx.get().secrets.get("KANDJI_SECRET")
    api_token = secret["api_token"]
    api_url = secret["api_url"]

    with get_kandji_client(api_url, api_token) as client:
        offset = 0
        max_limit_per_page = 300

        apps = []

        while True:
            response = client.get(
                url=f"/devices/{device_id}/apps",
                params={"limit": max_limit_per_page, "offset": offset},
            )
            response.raise_for_status()

            new_apps = response.json()["apps"]
            apps.extend(new_apps)

            offset += len(new_apps)
            if len(new_apps) < max_limit_per_page:
                break

        return apps
